{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Patterns.Completeness where

import Prelude hiding (print, putStrLn, putStr)
import IO.Output
import Text.PrettyPrint

import Control.Monad
import Control.Monad.Error
import Data.Map as Map
import Data.Set as Set
import Data.List as List
import Data.Maybe

import Syntax.Name
import Syntax.Values

import TypeChecker.Monad
import TypeChecker.Monad.Datatypes
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Errors
import TypeChecker.Abstractions
import TypeChecker.Values
import TypeChecker.UnfoldConstants

import Compilers.Run

import Utils.Monads
import Utils.Names
import Utils.Pretty
import Utils.Error


checkingClause trace = fsep $ pwords "checking clause " ++
			List.map (prettyPrec 100) (traceToPat trace)

data ValidConstructors = NotADatatype
		       | Valid [(Name, [(Maybe Value, Type)])]
    deriving (Show)

data UniView = UCon Name [Value]
	     | UArg Int
	     | UPair Value Value
	     | UOther Value

uniView :: Int -> Map.Map Int Value -> Value -> UniView
uniView fv s v
    | VAVCon x _ _ _ vs <- vAppView v = UCon x vs
    | VBId n <- vv, n < fv	    = 
	case Map.lookup n s of
	    Nothing	-> UArg n
	    Just v	-> uniView fv s v
    | VPair v1 v2 <- vv		    = UPair v1 v2
    | otherwise			    = UOther v
    where
	vv = vView v

telescope :: Type -> TC (TeleList Type)
telescope t =
    do	v <- unfoldConstants $ unEl t
	case vView v of
	    VPi w f ->
		do  TeleList tel <- telescope (absBody f)
		    return $ TeleList $ w : tel
	    _ -> return $ TeleList []

freshVal = vId <$> fresh "_"

-- | Compute the constructors that can appear in a given type.
validConstructors :: Maybe Value -> Type -> TC ValidConstructors
validConstructors i t =
    do	v <- unfoldConstants $ unEl t
	verbose 12 $ liftIO $ putStrLn $ "checking constructors of " ++
			case i of
			    Nothing -> show t
			    Just v  -> show v ++ " : " ++ show t
	r <- case vAppView v of
	    VAVDef d _ vs   ->
		do  mcons <- (Just . dataConstructors <$> lookupDatatype d) `catchError` \_ -> return Nothing
		    case mcons of
			Nothing   -> return NotADatatype
			Just cons ->
			    do	verbose 15 $ liftIO $ putStrLn $ show v ++ " is a datatype with constructors " ++ show cons
				let mi = case fmap vAppView i of
					    Just (VAVCon c _ _ _ vs) -> Just (c,vs)
					    _			     -> Nothing
				    cons' = case mi of
						Nothing	    -> List.map (\ (c,x) -> (c,(x,Nothing))) cons
						Just (c,ws) ->
						    Map.toList
						    $ Map.intersectionWith (,)
							(Map.fromList cons)
							(Map.singleton c $ Just ws)
				Valid . concat <$> mapM (uncurry $ valid vs) cons'
	    _		    -> return NotADatatype
	verbose 15 $ liftIO $ putStrLn $ "constructors: " ++ show r
	return r
    where
	-- We can only distinguish applications of different constructors.
	unify :: Int -> [Value] -> [Value] -> TC (Maybe (Map.Map Int Value))
	unify fv vs ws = uni Map.empty vs ws
	    where
		uni s [] [] =
		    do	verbose 10 $ liftIO $ putStrLn $ "Unified{" ++ show fv ++ "} " ++ show vs ++ " and " ++ show ws ++ ": " ++ show s
			return $ Just s
		uni s (v:vs) (w:ws) =
		    do	v' <- unfoldConstants v
			w' <- unfoldConstants w
			case (uniView fv s v', uniView fv s w') of
			    (UCon x vs', UCon y ws')
				| x /= y    -> return Nothing
				| otherwise -> uni s (vs' ++ vs) (ws' ++ ws)
			    (UArg n, UArg m)
				| n == m	-> uni s vs ws
			    (UArg n, _)		-> ifM (occurs n w')
						       (return Nothing)
						       (uni (Map.insert n w' s) vs ws)
			    (_, UArg m)		-> ifM (occurs m v')
						       (return Nothing)
						       (uni (Map.insert m v' s) vs ws)
			    (UPair v1 v2, _)	-> uni s (v1:v2:vs) (w1:w2:ws)
				where
				    w1 = vProject w' VFst
				    w2 = vProject w' VSnd
			    (_,UPair w1 w2)	-> uni s (v1:v2:vs) (w1:w2:ws)
				where
				    v1 = vProject v' VFst
				    v2 = vProject v' VSnd
			    (_, _)		-> uni s vs ws
		uni s _ _   = failure $ customError $ "impossible: TypeChecker.Patterns.Completeness.unify"

		occurs n v = Set.member n <$> foldValue Set.unions bind (fv . vView) v
		    where
			bind vs = Set.map (\x -> x - 1) $ Set.delete 0 vs
			fv (VBId n) = return $ Set.singleton n
			fv _	    = return $ Set.empty

	valid :: [Value] -> Name -> (DefValue, Maybe [Value]) -> TC [(Name, [(Maybe Value, Type)])] -- singleton or empty
	valid us con (DefVal _ _ _ pa _ t, inst) =
	    do	let t' = instantiate t $ reverse $ take pa us
		verbose 16 $ liftIO $ putStrLn $ "valid? " ++ show us ++ " -> " ++ show con ++ " : " ++ show t'
		(args, pars) <- conTel t'
		verbose 16 $ liftIO $
		    do	putStrLn $ "  args = " ++ show args
			putStrLn $ "  pars = " ++ show pars
		ms <- unify (length args) us pars -- (raiseBy fv pars) TODO: bug?
		verbose 16 $ liftIO $
		    case ms of
			Nothing	-> putStrLn "not valid"
			_	-> return ()
		case ms of
		    Nothing -> return []
		    Just s  ->
			do  inst0 <- case inst of
					Nothing	    -> return inst'
					Just inst0  ->
					    do	uncurry (unify 0) $ unzip
						    [ (i0, i1) | (i0, Just i1) <- zip inst0 inst' ]
						return $ List.map Just inst0
			    return [(con, zip inst0 args)]
			where
			    inst' = reverse $ List.map (flip Map.lookup s) [0 .. length args - 1]
	    where
		conTel t =
		    do	v <- unfoldConstants $ unEl t
			case vAppView v of
			    VAVPi w f     ->
				do  (tel, vs) <- conTel $ absBody f
				    return (w:tel, vs)
			    VAVDef d _ vs -> return ([], vs)
			    _	      -> failure $ customError "impossible: TypeChecker.Patterns.Completeness.conTel"

-- | Checks if any type in the telescope is empty.
--   @isEmptyTelescope _ _ = return False@
isEmptyTelescope :: [Maybe Value] -> TeleList Type -> TC Bool
isEmptyTelescope inst (TeleList tel) =
    do	d <- completenessDepth
	verbose 14 $ liftIO $ print $ sep [ text "isEmptyTelescope"
					  , nest 2 $ pretty $ zipWith pr inst tel
					  ]
	let isEmpty' d = isEmpty inst (List.map ((,) d) tel)
	--anyM isEmpty' [1..d] memory leak here!
	isEmpty' d
    where
	anyM f [] = return False
	anyM f (x:xs) =
	    do	b <- f x
		if b then return True
		     else anyM f xs

	pr Nothing t  = text "_" <+> text ":" <+> pretty t
	pr (Just v) t = pretty v <+> text ":" <+> pretty t

	isEmpty inst []		     = return False
	isEmpty (_:inst) ((0,t):tel) = isEmpty inst tel
	isEmpty (i:inst) ((d,t):tel) =
	    do	verbose 15 $ liftIO $ print $ sep [ text "isEmpty" <+> pretty d
						  , nest 2 $ pretty t
						  ]
		mcons <- validConstructors i t
		case mcons of
		    NotADatatype -> isEmpty inst tel
		    Valid []	 -> return True
		    Valid [con]	 -> isEmptyCon True con
		    Valid cons	 ->
			-- avoid exponential blow-up
			do  em1 <- or <$> mapM (isEmptyCon False) cons
			    if em1
				then return True
				else do	v <- maybe freshVal return i
					let (ds,tel')	   = unzip tel
					    TeleList tel'' = instantiate (TeleList tel') [v]
					isEmpty inst $ zip ds tel''
	    where
		isEmptyCon deep (c, argis) =
		    do  (_,ps,_,_) <- getDatatype $ unEl t
			let (is,args)	   = unzip argis
			    ar		   = length args
			    TeleList args' = instantiate (TeleList args) ps
			    vs		   = List.map vBId [ar - 1, ar - 2 .. 0]
			    v		   = vCon c ps 0 ar `vApps` vs
			    (ds,tel')	   = unzip tel
			    TeleList tel'' = instantiate (TeleList tel') [v]
			    argtel	   = zip (repeat (d - 1)) args'
			    tel''' | deep  = argtel ++ zip ds tel''
				   | True  = argtel
			    inst'  | deep  = is ++ inst
				   | True  = is
			isEmpty inst' tel'''
	isEmpty _ _ = failure $ customError "impossible: TypeChecker.Patterns.Completeness.isEmpty"

data Trace = TAny | TCon Name Int | TPair

traceToPat :: [Trace] -> [PatternValue]
traceToPat = ttp . reverse
    where
	ttp []		     = []
	ttp (TAny : ts)	     = VWildP : ttp ts
	ttp (TCon c ar : ts) = VConP c ps0 : ps1
	    where
		ps = ttp ts
		n  = length ps
		(ps0,ps1)
		    | n >= ar	= splitAt ar ps
		    | otherwise	= (ps ++ replicate (ar - n) VAbsurdP, [])
		    
	ttp (TPair : ts)     = VPairP p q : ps
	    where
		(p,q,ps) = case ttp ts of
			    p:q:ps  -> (p,q,ps)
			    [p]	    -> (p,VAbsurdP,[])
			    []	    -> (VAbsurdP,VAbsurdP,[])

type CaseTree = ClauseTree ()

-- TODO: Collect missing cases and give nice errors! Store ranges in compiled trees.
-- | Ensures that a definition is complete.
patternCompleteness :: Name -> DefValue -> TC ()
patternCompleteness name d@(DefVal _ _ _ _ def t)
    | List.null clauses = return ()	-- axiom
    | otherwise			   =
	sayWhat incompletePatterns $
	do  tree <- sayWhat (fsep $ pwords $ "failed to compile definition of " ++ show name)
		    $ compileClauses (\_ _ -> return ()) clauses
	    ar <- nArgs t
	    complete [] tree (replicate ar Nothing) t
    where
	clauses = getClausesRT def

	nArgs t =
	    do	v <- unfoldConstants $ unEl t
		case vView v of
		    VPi _ f -> (1 +) <$> nArgs (absBody f)
		    _	    -> return 0

	unfoldInst = maybe (return Nothing)
			   (\x -> Just <$> unfoldConstants x)

	complete :: [Trace] -> CaseTree -> [Maybe Value] -> Type -> TC ()
	complete trace tree inst t =
	    do	verbose 15 $ liftIO $ putStrLn $ "complete (" ++ show t ++ ")\n" ++
						 "  " ++ show tree ++ "\n" ++
						 "  " ++ show inst
		complete' trace tree inst t

	complete' _ (Done _) _ _	 = return ()

	complete' _ _ [] _		  = failure $ customError $ "impossible: TypeChecker.Patterns.Completeness.complete: empty instantiation"

	complete' trace (Bind tree) (i:inst) t  =
	    do	(w,_) <- isPi t
		i <- unfoldInst i
		em <- isEmptyTelescope [i] (TeleList [w])	-- short-cut
		if em then return ()
		      else do v <- maybe freshVal return i
			      complete (TAny : trace) tree inst =<< piApp t v

	complete' trace (Split tree) (i:inst) t =
	    do	(v,_) <- isPi t
		i <- unfoldInst i
		let is = case fmap vView i of
			    Just (VPair v w) -> [Just v, Just w]
			    _		     -> [Nothing, Nothing]
		v' <- unfoldConstants $ unEl v
		case vView v' of
		    VSigma w f	->
			do  t' <- piApp (raiseBy 2 t) (vPair (vBId 1) (vBId 0))
			    complete (TPair : trace) tree (is ++ inst) $ piAbstraction [w,absBody f] t'
		    _		-> failure $ customError $ "impossible: TypeChecker.Patterns.Completeness.complete: Non-sigma " ++ show v

	complete' trace (Case m Nothing) (i:_) t
	    | Map.null m =
		do  (w,f) <- isPi t
		    i <- unfoldInst i
		    em    <- isEmptyTelescope [i] (TeleList [w])
		    case em of
			True  -> return ()
			False ->
			    sayWhen (checkingClause trace)
			    $ failure $ customError $ show w ++ " is not (trivially) empty"
	complete' trace (Case m ca) (i:inst) t =
	    do	(w,f) <- isPi t
		i     <- unfoldInst i
		vcons <- validConstructors i w
		cons  <- Map.fromList <$> isDatatype w vcons
		let dif	= Map.differenceWith (\_ _ -> Nothing) cons m
		-- If there is a catch-all, expand it to all constructors that
		-- haven't been matched.
		m' <- case ca of
		    Nothing | Map.null dif  -> return Map.empty    -- complete!
			    | otherwise	    ->
				do  mapM_ (anyLaterEmptyArguments trace inst w f) $ Map.toList dif
				    return Map.empty
		    Just tree
			| Map.null dif	-> return Map.empty	-- the catch-all is redundant
			| otherwise	->
			    do	let m' = Map.mapWithKey (expanded tree) dif
				verbose 14 $ liftIO $ print $ sep [ text "expanded catch-all:"
								  , nest 2 $ vcat $ List.map pr $ Map.toList m'
								  ]
				return m'
			where pr (c,(_,(_,t))) = sep [ pretty c <+> text "->"
						     , nest 2 $ pretty t
						     ]
		let mWithArgs = intersectionWith (,) cons m `Map.union` m'
		mapM_ (completeCon trace inst w f) $ Map.toList mWithArgs

	expanded :: CaseTree -> Name -> [(Maybe Value, Type)] -> ([(Maybe Value, Type)], (Int, CaseTree))
	expanded t con argis = (argis, (n, bind n t))
	    where
		n = length argis

		bind 0 t = t
		bind n t = Bind $ bind (n - 1) t

	isDatatype t NotADatatype = failure $ customError $ show t ++ " is not a datatype"
	isDatatype _ (Valid cs)	= return cs

	anyLaterEmptyArguments trace inst w f (c,argis) =
	    do	(_,ps,_,_) <- getDatatype $ unEl w
		let (is,args)	   = unzip argis
		    ar		   = length is
		    vs		   = List.map vBId [ar - 1, ar - 2 .. 0]
		    TeleList args' = instantiate (TeleList args) ps
		    t'		   = piAbstraction args' $ f `absApp` (vCon c ps 0 ar `vApps` vs)
		verbose 15 $ liftIO $ print $ sep [ text "anyEmpty"
						  , nest 2 $ pretty t'
						  ]
		tel <- telescope t'
		em  <- isEmptyTelescope (is ++ inst) tel
		if em
		    then return ()
		    else sayWhen (checkingClause trace) $
			 failure $ customError $ "a case for " ++ show c ++ " is missing"

	isPi t =
	    do	v <- unfoldConstants $ unEl t
		case vView v of
		    VPi w f -> return (w,f)
		    _	    -> failure $ customError $ "impossible: TypeChecker.Patterns.Completeness.complete: Non-pi " ++ show t

	completeCon trace inst w f (con, (argis, (ar, tree))) =
	    do	verbose 15 $ liftIO $ putStrLn $ "completeCon [" ++ show (vPi w f) ++ "] " ++ show con ++ " " ++ show argis
		(_,ps,_,_) <- getDatatype $ unEl w
		let (is,args)	   = unzip argis
		    TeleList args' = instantiate (TeleList args) ps
		    t'		   = raiseBy ar f `absApp`
					(vCon con ps 0 ar `vApps`
					    List.map vBId [ar - 1,ar - 2..0]
					)
		verbose 15 $ liftIO $ putStrLn $ "  t'    = " ++ show t'
		verbose 15 $ liftIO $ putStrLn $ "  tree  = " ++ show tree
		verbose 15 $ liftIO $ putStrLn $ "  args' = " ++ show args'
		complete (TCon con ar:trace) tree (is ++ inst) (piAbstraction args' t')

